<?php
// This file is part of Moodle - http://moodle.org/
//
// Moodle is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Moodle is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Moodle.  If not, see <http://www.gnu.org/licenses/>.

/**
 * Contains the definiton of the pageone message processors (sends messages to users via text)
 *
 * @package   message_pageone
 * @copyright 2015 Tim Williams and TDM
 * @license   http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */


global $CFG;
require_once($CFG->dirroot.'/message/output/lib.php');
require_once($CFG->dirroot.'/blocks/pageone/pageonelib.php');

/**
 * Contains the definiton of the pageone message processors (sends messages to users via text)
 *
 * @package   message_pageone
 * @copyright 2015 Tim Williams and TDM
 * @license   http://www.gnu.org/copyleft/gpl.html GNU GPL v3 or later
 */
class message_output_pageone extends message_output {
    /**
     * Processes the message (sends by text).
     * @param object $eventdata the event data submitted by the message sender plus $eventdata->savedmessageid
     */
    public function send_message($eventdata) {

        // Does the user have a valid phone number? pageone_sendtext_expects this to already be the case.
        if (!pageone_has_valid_mobile_number($eventdata->userto)) {
            return false;
        }

        // We might want this to be true in some cases, but I've not found any yet, so this is for future use.
        // Note: Moodle Messages include sender in subject so not needed for those.

        $includesender = false;

        // For some event types the subject and small message are the same,
        // so check this to avoid a confusing "double" message.
        if ($eventdata->subject == $eventdata->smallmessage) {
            $subject = "";
        } else {
            $subject = $eventdata->subject;
        }

        $result = pageone_send_text(array($eventdata->userto), $eventdata->userfrom, $subject,
            $eventdata->smallmessage, $includesender, true);

        self::add_to_log($result->id, $this->find_course($eventdata), $eventdata->userfrom->id, $eventdata->userto->id,
            $subject, $eventdata->smallmessage, $includesender);

        return $result->ok;
    }

    /**
     * Adds a message to the PageOne message log table
     * @param int $resultid The result code sent by the PageOne Oventus server
     * @param int $course The Course ID this message relates to
     * @param int $userfrom The user id this message was sent from
     * @param int $userto The user id this message was sent to
     * @param string $subject The subject of the message
     * @param string $smallmessage The message text
     * @param bool $includesender True if the sender name was included in the message text
     **/
    private static function add_to_log($resultid, $course, $userfrom, $userto, $subject, $smallmessage, $includesender) {
        // Some colleges need all messages to be logged for accounting purposes, so if the config
        // setting is on, the messages will get logged.
        if (pageone_getconfigparam("message_pageone", "log") == 1) {
            $log = new stdClass;
            $log->ovid       = $resultid;
            $log->courseid   = $course;
            $log->userid     = $userfrom;
            $log->mailto     = $userto;
            $log->subject    = $subject;
            $log->message    = $smallmessage;
            $log->attachment = 0;
            $log->format     = FORMAT_PLAIN;
            $log->timesent   = time();
            $log->messagetype = TYPE_TEXT_AUTO;
            $log->includefrom = $includesender;
            $log->failednumbers = '';

            global $DB;
            $DB->insert_record('block_pageone_log', $log);
        }
    }

    /**
     * Finds the course id which is associated with the supplied event
     * @param object $eventdata A Moodle event object
     * @return The course ID
     **/
    private function find_course($eventdata) {

        // If this is a course module.
        if (strpos($eventdata->component, "mod_") == 0 && property_exists($eventdata, "contecturl")) {
            // Moodle doesn't give us a context or course ID in this case, but we can (probably) get one from the contexturl.
            $start = strpos($eventdata->contexturl, "id=");
            if ($start > -1) {
                $start = start + 3;
                $end = strpos($eventdata->contexturl, "&", $start);
                if ($end > $start) {
                    $contid = substr($eventdata->contexturl, $start, $end - $start);
                } else {
                    $contid = substr($eventdata->contexturl, $start);
                }

                global $DB;
                $mod = $DB->get_record("course_modules", array('id' => $contid));
                return $mod->course;
            }
        }

        // This is might be different for other event types, but I haven't found this to be the case yet.

        // Try $COURSE.
        global $COURSE;
        if (isset($COURSE->id) && $COURSE->id > 0) {
            return $COURSE->id;
        }

        // Default to course 1 if we can't work out what course it should be.
        return 1;
    }

    /**
     * Creates necessary fields in the messaging config form.
     *
     * @param array $preferences An array of user preferences
     */
    public function config_form($preferences) {
        if (!pageone_is_configured()) {
            return "";
        }

        $opts = "";

        if (pageone_getconfigparam("message_pageone", "assign_deadline")) {
            $opts .= $this->get_option($preferences, "assign_deadline");
        }

        if (pageone_getconfigparam("message_pageone", "forum")) {
            $opts .= $this->get_option($preferences, "forum");
        }

        return $opts;
    }

    /**
     * Gets an HTML inpuct checkbox for the specified iption
     * @param object $preferences The preferences object to check
     * @param string $opt The option name
     * @return An HTML checkbox
     **/
    private function get_option($preferences, $opt) {
        $k = "pageone_" . $opt;

        if ($preferences->$k == 1) {
            $c = "checked='checked'";
        } else {
            $c = "";
        }

        return "<p><input type='checkbox' name='pageone_" . $opt . "' value='1' " . $c . " />".
            get_string("user_".$opt, "message_pageone") . "</p>";
    }

    /**
     * Parses the submitted form data and saves it into preferences array.
     *
     * @param stdClass $form preferences form class
     * @param array $preferences preferences array
     */
    public function process_form($form, &$preferences) {
        if (isset($form->pageone_assign_deadline)) {
            $preferences['message_processor_pageone_assign_deadline'] = $form->pageone_assign_deadline;
        } else {
            $preferences['message_processor_pageone_assign_deadline'] = 0;
        }

        if (isset($form->pageone_forum)) {
            $preferences['message_processor_pageone_forum'] = $form->pageone_forum;
        } else {
            $preferences['message_processor_pageone_forum'] = 0;
        }
    }

    /**
     * Loads the config data from database to put on the form during initial form display
     *
     * @param array $preferences preferences array
     * @param int $userid the user id
     */
    public function load_data(&$preferences, $userid) {
        $preferences->pageone_assign_deadline = get_user_preferences('message_processor_pageone_assign_deadline', '', $userid);
        if ($preferences->pageone_assign_deadline == null) {
            $preferences->pageone_assign_deadline = pageone_getconfigparam("message_pageone", "assign_deadline_user");
        }

        $preferences->pageone_forum = get_user_preferences('message_processor_pageone_forum', '', $userid);
        if ($preferences->pageone_forum == null) {
            $preferences->pageone_forum = pageone_getconfigparam("message_pageone", "forum_user");
        }
    }

    /**
     * Returns true as message can be sent to internal support user.
     *
     * @return bool
     */
    public function can_send_to_any_users() {
        return false;
    }

    /**
     * Returns the message processors default settings. MESSAGE_DISALLOWED is the default for this plugin.
     *
     * @return int The Default message output settings expressed as a bit mask
     *         MESSAGE_DEFAULT_LOGGEDIN + MESSAGE_DEFAULT_LOGGEDOFF + MESSAGE_DISALLOWED|MESSAGE_PERMITTED|MESSAGE_FORCED
     */
    public function get_default_messaging_settings() {
        return MESSAGE_DISALLOWED;
    }

    /**
     * Cron job functions
     **/
    public function cron() {
        // Skip this if PageOne isn't configured.
        if (!pageone_is_configured()) {
            return true;
        }

        $cronhour = pageone_getconfigparam("message_pageone", "cron_time");
        $assigndeadline = pageone_getconfigparam("message_pageone", "assign_deadline");

        // Exit if messaging output settings are missing.
        if ($cronhour == null || $assigndeadline == null) {
            return true;
        }

        // Exit if deadline digests are turned off.
        if (!$assigndeadline) {
            return true;
        }

        $now = new DateTime();
        $nowhour = $now->format("H");

        // If the current time is before the time the job is supposed to run at, then skip the job.
        if ($nowhour < $cronhour) {
            return true;
        }

        $lastcron = pageone_getconfigparam("message_pageone", "lastcron");

        if ($lastcron == null) {
            $this->do_cron();
        } else {
            // Check to see if the cron job has already been run today.
            $lc = new DateTime('@' . $lastcron);
            if ($lc->format('Y-m-d') != $now->format('Y-m-d')) {
                $this->do_cron();
            }
        }
        return true;
    }

    /**
     * Executes the cron functions if cron job is needed
     **/
    private function do_cron() {
        set_config("lastcron", time(), "message_pageone");

        global $DB, $CFG;

        $deadlinetime = intval(pageone_getconfigparam("message_pageone", "deadline_days")) * 24 * 60 * 60;
        $assignmoduleid = $DB->get_record("modules", array("name" => "assign"))->id;
        $messagedefault = pageone_getconfigparam("message_pageone", "assign_deadline_user");
        $messagetext = get_string("assign_deadline_text", "message_pageone",
            pageone_getconfigparam("message_pageone", "deadline_days"));

        // Find all the assignments due within the specified period.
        $now = time();
        $later = $now + $deadlinetime;

        $assigns = $DB->get_recordset_select("assign", "duedate > ".$now." AND duedate < ".$later);
        //echo "MoodleMobile digest messages from ".date_format(new DateTime("@".$now), 'Y-m-d H:i:s')." (".$now.") to ".date_format(new DateTime("@".$later),'Y-m-d H:i:s')." (".$later."): ".count($assigns)."\n";
        $messages = array();
        $allusers = array();
        // Loop through all the assignments.
        foreach ($assigns as $assign) {
            // Check the database to make sure we haven't done this assignment already.
            // Also if the due dates no longer match, then assume we need to re-send the notification.
            $check = $DB->get_record("message_pageone_assign", array("assignid" => $assign->id));
            if ($check && $assign->duedate <= $check->duedate) {
                continue;
            }

            // Add/Update the check record to prevent duplicate messages.
            if (!$check) {
                $check = new stdClass;
                $check->assignid = $assign->id;
                $check->duedate = $assign->duedate;
                $check->duedate = $assign->duedate;
                $DB->insert_record("message_pageone_assign", $check);
            } else {
                $check->duedate = $assign->duedate;
                $DB->update_record("message_pageone_assign", $check);
            }

            // Get the course module.
            $cm = $DB->get_record("course_modules", array("module" => $assignmoduleid, "course" => $assign->course,
                "instance" => $assign->id));
            $course = $DB->get_record("course", array("id" => $assign->course));
            $context = context_module::instance($cm->id);

            // Find all the students who have the ability to submit to this assignment.
            $users = get_users_by_capability($context, "mod/assign:submit");
            foreach ($users as $user) {

                // Does this user want notifications?
                if ($this->send_assign_deadlines($user->id, $messagedefault)) {
                    $allusers[$user->id] = $user;

                    if (array_key_exists($user->id, $messages)) {
                        $messages[$user->id] .= "\n".$course->shortname.",".$assign->name;
                    } else {
                        $messages[$user->id] = $course->shortname.",".$assign->name;
                    }

                }
            }
        }

        $adminuser = $DB->get_record("user", array("username" => "admin"));

        foreach ($messages as $userid => $message) {
            echo "Texting user=" . $allusers[$userid]->firstname . " " . $allusers[$userid]->lastname . "\n";

            $result = pageone_send_text(array($allusers[$userid]), $adminuser, $messagetext, $message, false, true);
            self::add_to_log($result->id, $course->id, $adminuser->id, $allusers[$userid]->id,
                $messagetext, $message, false);
        }
    }

    /**
     * Gets the assignment deadline notification preference
     * @param int $userid The ID of the user to check
     * @param bool $messagedefault The default value (true or false) to use if the user has no preference
     * @return The preference (true or false)
     **/
    private function send_assign_deadlines($userid, $messagedefault) {
        global $DB;

        $prefs = $DB->get_record("user_preferences",
            array("userid" => $userid, "name" => "message_processor_pageone_assign_deadline"));
        if ($prefs) {
            if ($prefs->value = "1") {
                return true;
            } else {
                return false;
            }
        }

        // No preference, use the default.
        return $messagedefault;
    }

    /**
     * Gets the forum notification preference
     * @param int $userid The ID of the user to check
     * @param bool $forumdefault The default value (true or false) to use if the user has no preference
     * @return The preference (true or false)
     **/
    private static function send_forum_notify($userid, $forumdefault) {
        global $DB;

        $prefs = $DB->get_record("user_preferences", array("userid" => $userid, "name" => "message_processor_pageone_forum"));
        if ($prefs) {
            if ($prefs->value = "1") {
                return true;
            } else {
                return false;
            }
        }

        // No preference, use the default.
        return $forumdefault;
    }

    /**
     * Deals with a forum post event
     * @param object $event The forum event to process 
     **/
    public static function forum_post($event) {
        global $DB, $USER;
        $adminuser = $DB->get_record("user", array("username" => "admin"));
        $forumdefault = pageone_getconfigparam("message_pageone", "forum_user");

        $post = $DB->get_record("forum_posts", array("id" => $event->objectid));
        $disc = $DB->get_record("forum_discussions", array("id" => $post->discussion));
        $forum = $DB->get_record("forum", array("id" => $disc->forum));
        $course = $DB->get_record("course", array("id" => $disc->course));

        $text = new stdClass;
        $text->sub = $post->subject;
        $text->fname = $forum->name;
        $text->cname = $course->shortname;
        $message = get_string("forum_reply", "message_pageone", $text);

        $userlist = array();
        self::get_user_list($post, $forumdefault, $userlist);
        foreach ($userlist as $userid => $postid) {
            if ($userid == $USER->id) {
                continue;
            }

            $user = $DB->get_record("user", array("id" => $userid));
            $result = pageone_send_text(array($user), $adminuser, "", $message, false, true);
            $DB->insert_record("message_pageone_forum", array("userid" => $userid, "postid" => $postid));
            if (isset($result->id))
                self::add_to_log($result->id, $course->id, $adminuser->id, $userid, "", $message, false);
        }
    }

    /**
     * Gets all the users who are associated with a specific forum posting
     * @param object $post The post to check
     * @param bool $forumdefault The system notification default for forum notifications
     * @param array $userlist Pass an array here, it will be populated by this method with user ID's
     **/
    private static function get_user_list($post, $forumdefault, &$userlist) {
        global $DB;
        $alreadysent = $DB->get_record("message_pageone_forum", array("userid" => $post->userid, "postid" => $post->id));
        $usersend = self::send_forum_notify($post->userid, $forumdefault);
        $already = array_key_exists($post->userid, $userlist);

        if (!$alreadysent && self::send_forum_notify($post->userid, $forumdefault) && !array_key_exists($post->userid, $userlist)) {
            $userlist[$post->userid] = $post->id;
        }

        if ($post->parent == 0) {
            return;
        }

        $nextpost = $DB->get_record("forum_posts", array("id" => $post->parent));
        self::get_user_list($nextpost, $forumdefault, $userlist);
    }
}
